from typing import Any, Dict, Optional

from typing_extensions import Literal

from whylogs.api.logger.logger import Logger
from whylogs.api.logger.result_set import ResultSet
from whylogs.api.logger.rolling import TimedRollingLogger
from whylogs.api.logger.transient import TransientLogger
from whylogs.core import DatasetProfile, DatasetSchema
from whylogs.core.stubs import pd


def log(
    obj: Any = None,
    *,
    pandas: Optional[pd.DataFrame] = None,
    row: Optional[Dict[str, Any]] = None,
    schema: Optional[DatasetSchema] = None,
) -> ResultSet:
    return TransientLogger(schema=schema).log(obj, pandas=pandas, row=row)


def read(path: str) -> ResultSet:
    return ResultSet.read(path)


def write(profile: DatasetProfile, base_dir: str) -> None:
    return profile.write(base_dir)


def logger(
    schema: Optional[DatasetSchema] = None,
    *,
    mode: Literal["transient", "rolling"] = "transient",
    **kwargs: Any,
) -> Logger:
    """

    Returns:
        Logger: a logger object
    """
    if mode == "transient":
        return TransientLogger(schema)
    elif mode == "rolling":
        return TimedRollingLogger(schema, **kwargs)
    else:
        raise ValueError("Unsupported logger mode")


__ALL__ = [
    "log",
    "read",
    "write",
    "logger",
    "Logger",
    "ResultSet",
    "RollingLogger",
]



import dataclasses
import math
import statistics
import sys
from abc import ABC, abstractmethod
from dataclasses import asdict, dataclass
from typing import Any, Dict, List, Tuple, Type, TypeVar, Union

import whylogs_sketching as ds  # type: ignore
from google.protobuf.struct_pb2 import Struct
from typing_extensions import TypeAlias

from whylogs.core.configs import SummaryConfig
from whylogs.core.metrics.maths import (
    VarianceM2Result,
    parallel_variance_m2,
    welford_online_variance_m2,
)
from whylogs.core.metrics.metric_components import (
    FractionalComponent,
    FrequentStringsComponent,
    HllComponent,
    KllComponent,
    MaxIntegralComponent,
    MetricComponent,
    MinIntegralComponent,
)
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricComponentMessage, MetricMessage

T = TypeVar("T")
M = TypeVar("M", bound=MetricComponent)
NUM = TypeVar("NUM", float, int)

METRIC = TypeVar("METRIC", bound="Metric")
ColumnSchema: TypeAlias = "ColumnSchema"  # type: ignore


_METRIC_DESERIALIZER_REGISTRY: Dict[str, Type[METRIC]] = {}  # type: ignore


def custom_metric(metric: Type[METRIC], schema=None) -> Type[METRIC]:  # type: ignore
    global _METRIC_DESERIALIZER_REGISTRY
    _METRIC_DESERIALIZER_REGISTRY[metric.get_namespace(schema)] = metric
    return metric


@dataclass(frozen=True)
class OperationResult:
    failures: int = 0
    successes: int = 0

    @classmethod
    def ok(cls, cnt: int = 1) -> "OperationResult":
        return OperationResult(successes=cnt, failures=0)

    @classmethod
    def failed(cls, cnt: int = 1) -> "OperationResult":
        return OperationResult(successes=0, failures=cnt)

    def __add__(self, other: "OperationResult") -> "OperationResult":
        return OperationResult(
            successes=self.successes + other.successes,
            failures=self.failures + other.failures,
        )


class Metric(ABC):
    @classmethod
    def get_namespace(cls, schema: ColumnSchema) -> str:
        return cls.zero(schema).namespace

    @property
    @abstractmethod
    def namespace(self) -> str:
        raise NotImplementedError

    def __post_init__(self):
        global _METRIC_DESERIALIZER_REGISTRY
        _METRIC_DESERIALIZER_REGISTRY[self.namespace] = self.__class__

    def __add__(self: METRIC, other: METRIC) -> METRIC:
        return self.merge(other)

    def merge(self: METRIC, other: METRIC) -> METRIC:
        res: Dict[str, MetricComponent] = {}
        for k, v in self.__dict__.items():
            if isinstance(v, MetricComponent):
                res[k] = v + other.__dict__[k]

        return self.__class__(**res)

    def to_protobuf(self) -> MetricMessage:
        if not dataclasses.is_dataclass(self):
            raise ValueError("Metric object is not a dataclass")

        res = {}
        for k, v in self.__dict__.items():
            if not isinstance(v, MetricComponent):
                continue
            res[k] = v.to_protobuf()
        return MetricMessage(metric_components=res)

    def get_component_paths(self) -> List[str]:
        res = []
        for k, v in self.__dict__.items():
            if not isinstance(v, MetricComponent):
                continue
            res.append(k)
        return res

    @abstractmethod
    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        raise NotImplementedError

    @abstractmethod
    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def zero(cls: Type[METRIC], schema: ColumnSchema) -> METRIC:
        pass

    @classmethod
    def from_protobuf(cls: Type[METRIC], msg: MetricMessage) -> METRIC:
        if not dataclasses.is_dataclass(cls):
            raise ValueError(f"Metric class: {cls} is not a dataclass")

        components: Dict[str, MetricComponent] = {}
        for k, m in msg.metric_components.items():
            components[k] = MetricComponent.from_protobuf(m)

        return cls(**components)


@dataclass(frozen=True)
class IntsMetric(Metric):
    max: MaxIntegralComponent
    min: MinIntegralComponent

    @property
    def namespace(self) -> str:
        return "ints"

    def columnar_update(self, data: PreprocessedColumn) -> OperationResult:
        if data.len <= 0:
            return OperationResult.ok(0)

        successes = 0

        max_ = self.max.value
        min_ = self.min.value
        if data.numpy.ints is not None:
            max_ = max([max_, data.numpy.ints.max()])
            min_ = min([min_, data.numpy.ints.min()])
            successes += len(data.numpy.ints)

        if data.list.ints is not None:
            l_max = max(data.list.ints)
            l_min = min(data.list.ints)
            max_ = max([max_, l_max])
            min_ = min([min_, l_min])
            successes += len(data.list.ints)

        self.max.set(max_)
        self.min.set(min_)
        return OperationResult.ok(successes)

    @classmethod
    def zero(cls, schema: ColumnSchema) -> "IntsMetric":
        return IntsMetric(max=MaxIntegralComponent(-sys.maxsize), min=MinIntegralComponent(sys.maxsize))

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Union[int, float, str, None]]:
        return {"max": self.max.value, "min": self.min.value}


@dataclass(frozen=True)
class DistributionMetric(Metric):
    kll: KllComponent
    mean: FractionalComponent
    m2: FractionalComponent

    @property
    def namespace(self) -> str:
        return "distribution"

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Union[int, float, str, None]]:
        if self.n == 0:
            quantiles = [None, None, None, None, None]
        else:
            quantiles = self.kll.value.get_quantiles([0.1, 0.25, 0.5, 0.75, 0.9])
        return {
            "mean": self.mean.value,
            "stddev": self.stddev,
            "n": self.kll.value.get_n(),
            "max": self.kll.value.get_max_value(),
            "min": self.kll.value.get_min_value(),
            "q_10": quantiles[0],
            "q_25": quantiles[1],
            "median": quantiles[2],
            "q_75": quantiles[3],
            "q_90": quantiles[4],
        }

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        """
        Update the operation

        Algorithm: https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Parallel_algorithm

        Args:
            view: the preprocessed column

        Returns:
            how many successful operations we had
        """
        successes = 0

        first = VarianceM2Result(n=self.kll.value.get_n(), mean=self.mean.value, m2=self.m2.value)

        if view.numpy.len > 0:
            for arr in [view.numpy.floats, view.numpy.ints]:
                if arr is not None:
                    self.kll.value.update(arr)
                    n_b = len(arr)
                    if n_b > 1:
                        n_b = len(arr)
                        mean_b = arr.mean()
                        m2_b = arr.var() * (n_b - 1)

                        second = VarianceM2Result(n=n_b, mean=mean_b, m2=m2_b)
                        first = parallel_variance_m2(first=first, second=second)
                    else:
                        # fall back to https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance
                        # #Weighted_incremental_algorithm
                        first = welford_online_variance_m2(existing=first, new_value=first[0])

        for lst in [view.list.ints, view.list.floats]:
            if lst is not None and len(lst) > 0:
                self.kll.value.update_list(lst)
                n_b = len(lst)
                if n_b > 1:
                    mean_b = statistics.mean(lst)
                    m2_b = statistics.pvariance(lst) * (n_b - 1)
                    second = VarianceM2Result(n=n_b, mean=mean_b, m2=m2_b)

                    first = parallel_variance_m2(first=first, second=second)
                else:
                    first = welford_online_variance_m2(existing=first, new_value=lst[0])

        self.mean.set(first.mean)
        self.m2.set(first.m2)

        return OperationResult.ok(successes)

    def merge(self, other: "DistributionMetric") -> "DistributionMetric":
        # calculate mean and m2
        a_n = self.kll.value.get_n()
        b_n = other.kll.value.get_n()

        delta = other.mean.value - self.mean.value
        new_n = a_n + b_n
        m2 = self.m2.value + other.m2.value + delta**2 * a_n * b_n / new_n

        mean = (a_n / new_n) * (self.mean.value) + (b_n / new_n) * (other.mean.value)

        # merge the sketch
        kll = self.kll + other.kll

        return DistributionMetric(kll=kll, mean=FractionalComponent(mean), m2=FractionalComponent(m2))

    @property
    def n(self) -> float:
        return self.kll.value.get_n()

    @property
    def variance(self) -> float:
        """Returns the population variance of the stream.

        https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance#Welford's_online_algorithm
        """
        if self.n <= 1:
            return float(0)
        return self.m2.value / (self.kll.value.get_n() - 1)

    @property
    def stddev(self) -> float:
        if self.n <= 1:
            return float(0)
        return math.sqrt(self.variance)

    @property
    def avg(self) -> float:
        return self.mean.value

    @property
    def max(self) -> float:
        return self.kll.value.get_max_value()

    @property
    def min(self) -> float:
        return self.kll.value.get_min_value()

    @classmethod
    def zero(cls, schema: ColumnSchema) -> "DistributionMetric":
        sk = ds.kll_doubles_sketch(k=schema.cfg.kll_k)
        return DistributionMetric(
            kll=KllComponent(sk),
            mean=FractionalComponent(0.0),
            m2=FractionalComponent(0.0),
        )


@dataclass(frozen=True)
class FrequentItem:
    value: str
    est: int
    upper: int
    lower: int


@dataclass(frozen=True)
class FrequentItemsMetric(Metric):
    frequent_strings: FrequentStringsComponent

    @property
    def namespace(self) -> str:
        return "frequent_items"

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        successes = 0
        for arr in [view.numpy.floats, view.numpy.ints]:
            if arr is not None:
                self.frequent_strings.value.update_np(arr)
                successes += len(arr)
        if view.pandas.strings is not None:
            self.frequent_strings.value.update_str_list(view.pandas.strings.to_list())
            successes += len(view.pandas.strings)

        if view.list.ints is not None:
            self.frequent_strings.value.update_int_list(view.list.ints)
            successes += len(view.list.ints)

        if view.list.floats is not None:
            self.frequent_strings.value.update_double_list(view.list.floats)
            successes += len(view.list.floats)

        if view.list.strings is not None:
            self.frequent_strings.value.update_str_list(view.list.strings)
            successes += len(view.list.strings)

        failures = 0
        if view.list.objs is not None:
            successes += len(view.list.objs)
        if view.pandas.objs is not None:
            successes += len(view.pandas.objs)

        return OperationResult(successes=successes, failures=failures)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        all_freq_items = self.frequent_strings.value.get_frequent_items(
            cfg.frequent_items_error_type.to_datasketches_type()
        )
        if cfg.frequent_items_limit > 0:
            limited_freq_items = all_freq_items[: cfg.frequent_items_limit]
        else:
            limited_freq_items = all_freq_items
        items = [FrequentItem(value=x[0], est=x[1], upper=x[2], lower=x[3]) for x in limited_freq_items]
        return {"frequent_strings": items}

    @classmethod
    def zero(cls, schema: ColumnSchema) -> "FrequentItemsMetric":
        sk = ds.frequent_strings_sketch(lg_max_k=schema.cfg.fi_lg_max_k)
        return FrequentItemsMetric(frequent_strings=FrequentStringsComponent(sk))


@dataclass(frozen=True)
class CardinalityMetric(Metric):
    hll: HllComponent

    @property
    def namespace(self) -> str:
        return "cardinality"

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        successes = 0
        if view.numpy.len > 0:
            if view.numpy.ints is not None:
                self.hll.value.update_np(view.numpy.ints)
                successes += len(view.numpy.ints)
            if view.numpy.floats is not None:
                self.hll.value.update_np(view.numpy.floats)
                successes += len(view.numpy.floats)
        if view.pandas.strings is not None:
            self.hll.value.update_str_list(view.pandas.strings.to_list())
            successes += len(view.pandas.strings)

        # update everything in the remaining lists
        if view.list.ints:
            self.hll.value.update_int_list(view.list.ints)
            successes += len(view.list.ints)
        if view.list.floats:
            self.hll.value.update_double_list(view.list.floats)
            successes += len(view.list.floats)
        if view.list.strings:
            self.hll.value.update_str_list(view.list.strings)
            successes += len(view.list.strings)

        failures = 0
        if view.list.objs:
            failures = len(view.list.objs)
        return OperationResult(successes=successes, failures=failures)

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return {
            "est": self.hll.value.get_estimate(),
            f"upper_{cfg.hll_stddev}": self.hll.value.get_upper_bound(cfg.hll_stddev),
            f"lower_{cfg.hll_stddev}": self.hll.value.get_lower_bound(cfg.hll_stddev),
        }

    @classmethod
    def zero(cls, schema: ColumnSchema) -> "CardinalityMetric":
        sk = ds.hll_sketch(schema.cfg.hll_lg_k)
        return CardinalityMetric(hll=HllComponent(sk))


def _drop_private_fields(data: List[Tuple[str, Any]]) -> Dict[str, Any]:
    return {k: v for k, v in data if not k.startswith("_")}


_STRUCT_NAME = "dataclass_param"


class CustomMetricBase(Metric, ABC):
    """
    You can use this as a base class for custom metrics that don't use
    the supplied or custom MetricComponents. Subclasses must be decorated with
    @dataclass. All fields not prefixed with an underscore will be included
    in the summary and will be [de]serialized. Such subclasses will need to
    implement the namespace, merge, and zero methods.
    """

    def get_component_paths(self) -> List[str]:
        return [_STRUCT_NAME]  # Assumes everything to be serde'd will be in the Struct

    def to_summary_dict(self, cfg: SummaryConfig) -> Dict[str, Any]:
        return asdict(self, dict_factory=_drop_private_fields)

    def to_protobuf(self) -> MetricMessage:
        mcm = MetricComponentMessage(dataclass_param=Struct())
        mcm.dataclass_param.update(asdict(self, dict_factory=_drop_private_fields))
        return MetricMessage(metric_components={_STRUCT_NAME: mcm})

    @classmethod
    def from_protobuf(cls: Type[METRIC], msg: MetricMessage) -> METRIC:
        my_struct = msg.metric_components[_STRUCT_NAME].dataclass_param
        return cls(**my_struct)



from abc import ABC, abstractmethod
from typing import Any

from whylogs.api.writer import Writer, Writers
from whylogs.core import DatasetProfile, DatasetProfileView


class ResultSet(ABC):
    """
    A holder object for profiling results.

    A whylogs.log call can result in more than one profile. This wrapper class
    simplifies the navigation among these profiles.

    Note that currently we only hold one profile but we're planning to add other
    kinds of profiles such as segmented profiles here.
    """

    @classmethod
    def read(cls, multi_profile_file: str) -> "ResultSet":
        # TODO: parse multiple profile
        view = DatasetProfileView.read(multi_profile_file)
        return ViewResultSet(view=view)

    def writer(self, name: str = "local") -> "ResultSetWriter":
        writer = Writers.get(name)
        return ResultSetWriter(results=self, writer=writer)

    @abstractmethod
    def view(self) -> DatasetProfileView:
        pass

    @abstractmethod
    def profile(self) -> DatasetProfile:
        pass


class ProfileResultSet(ResultSet):
    def __init__(self, profile: DatasetProfile) -> None:
        self._profile = profile

    def profile(self) -> DatasetProfile:
        return self._profile

    def view(self) -> DatasetProfileView:
        return self._profile.view()


class ViewResultSet(ResultSet):
    def __init__(self, view: DatasetProfileView) -> None:
        self._view = view

    def profile(self) -> DatasetProfile:
        raise ValueError("No profile available. Can only view")

    def view(self) -> DatasetProfileView:
        return self._view


class ResultSetWriter:
    """
    Result of a logging call.

    A result set might contain one or multiple profiles or profile views.
    """

    def __init__(self, results: ResultSet, writer: Writer):
        self._result_set = results
        self._writer = writer

    def option(self, **kwargs: Any) -> "ResultSetWriter":
        self._writer.option(**kwargs)
        return self

    def write(self, **kwargs) -> None:
        # TODO: multi-profile writer
        view = self._result_set.view()
        self._writer.write(profile=view, **kwargs)



from logging import getLogger
from typing import Any, Callable, List, Mapping, Optional, Tuple, Union

from whylogs.core.proto.v0 import Op
from whylogs.core.utils import get_distribution_metrics
from whylogs.core.view.dataset_profile_view import DatasetProfileView

logger = getLogger(__name__)

_summary_funcs1: Mapping[int, Callable[..., Callable[[Any], Any]]] = {
    # functions that compare a summary field to a literal value.
    Op.LT: lambda f, v: lambda s: getattr(s, f) < v,
    Op.LE: lambda f, v: lambda s: getattr(s, f) <= v,
    Op.EQ: lambda f, v: lambda s: getattr(s, f) == v,
    Op.NE: lambda f, v: lambda s: getattr(s, f) != v,
    Op.GE: lambda f, v: lambda s: getattr(s, f) >= v,
    Op.GT: lambda f, v: lambda s: getattr(s, f) > v,
    Op.BTWN: lambda f, v1, v2: lambda s: v1 <= getattr(s, f) <= v2,
    Op.IN: lambda f, v: lambda s: getattr(s, f) in v,
    Op.CONTAIN: lambda f, v: lambda s: v in getattr(s, f),
}


def _create_column_profile_summary_object(**kwargs: Optional[float]) -> type:
    """
    Wrapper method for summary constraints update object creation

    Parameters
    ----------
    number_summary : NumberSummary
        Summary object generated from NumberTracker
        Used to unpack the metrics as separate items in the dictionary
    kwargs : Summary objects or datasketches objects
        Used to update specific constraints that need additional calculations
    Returns
    -------
    Anonymous object containing all of the metrics as fields with their corresponding values
    """

    column_summary = {}

    column_summary.update(kwargs)

    return type("Object", (), column_summary)


class SummaryConstraint:
    def __init__(
        self,
        first_field: str,
        op: int,
        value: Optional[float] = None,
        second_field: str = None,
        name: str = None,
        verbose: bool = False,
    ) -> None:
        self.first_field = first_field
        self._verbose = verbose
        self._name = name
        self.op = op
        self.second_field = second_field
        self.value = value
        self.total = 0
        self.failures = 0

        if value is not None and second_field is None:
            # field-value summary comparison
            self.func = _summary_funcs1[op](first_field, value)

    def update(self, update_summary: object) -> None:
        constraint_type_str = "table shape" if self.first_field in ("columns", "total_row_number") else "summary"

        self.total += 1

        if not self.func(update_summary):
            self.failures += 1
            if self._verbose:
                logger.info(f"{constraint_type_str} constraint {self._name} failed")

    def report(self) -> Tuple[Any, ...]:
        return (self._name, self.total, self.failures)


class SummaryConstraints:
    def __init__(self, constraints: Union[List[SummaryConstraint], Mapping[str, SummaryConstraint]] = None):
        """
        SummaryConstraints is a container for multiple summary constraints,
        generally associated with a single ColumnProfile.

        Parameters
        ----------
        constraints : Mapping[str, SummaryConstraint]
            A dictionary of summary constraints with their names as keys.
            Can also accept a list of summary constraints.

        """

        if constraints is None:
            constraints = dict()

        # Support list of constraints for back compat with previous version.
        if isinstance(constraints, list):
            self.constraints = {constraint._name: constraint for constraint in constraints}
        elif isinstance(constraints, dict):
            self.constraints = constraints
        else:
            raise ValueError("Unexpected type for constraints parameter.")

    def __getitem__(self, name: str) -> Optional[SummaryConstraint]:
        if self.constraints:
            return self.constraints.get(name)
        return None

    def update(self, v: type) -> None:
        for c in self.constraints.values():
            c.update(v)

    def report(self) -> Optional[List[tuple]]:
        v = [c.report() for c in self.constraints.values()]
        if len(v) > 0:
            return v
        return None


class DatasetConstraints:
    def __init__(
        self,
        summary_constraints: Mapping[str, SummaryConstraints] = None,
    ):
        """
        DatasetConstraints is a container for multiple types of constraint containers, such as ValueConstraints,
        SummaryConstraints, and MultiColumnValueConstraints.
        Used for wrapping constraints that should be applied on a single data set.

        Parameters
        ----------
        summary_constraints : Mapping[str, SummaryConstraints]
            A dictionary where the keys correspond to the name of the feature for which the supplied value
            represents the SummaryConstraints to be executed

        """

        if summary_constraints is None:
            summary_constraints = dict()
        for k, v in summary_constraints.items():
            if isinstance(v, list):
                summary_constraints[k] = SummaryConstraints(v)
        self.summary_constraint_map = summary_constraints

    def __call__(self, profile_view: DatasetProfileView) -> List[Tuple[Any, ...]]:
        return self._constraints_report(profile_view)

    def _constraints_report(self, profile_view: DatasetProfileView) -> List[Tuple[Any, ...]]:
        if not isinstance(profile_view, DatasetProfileView):
            raise ValueError("Constraints must receive a DatasetProfileView object")
        columns = profile_view.get_columns()
        for k, v in self.summary_constraint_map.items():
            if isinstance(v, list):
                self.summary_constraint_map[k] = SummaryConstraints(v)
        for feature_name, constraints in self.summary_constraint_map.items():
            if feature_name in columns:
                min_val, max_val, range_val = get_distribution_metrics(columns[feature_name])
                update_obj = _create_column_profile_summary_object(max=max_val, min=min_val, range=range_val)
                constraints.update(update_obj)
            else:
                logger.debug(f"unkown feature '{feature_name}' in summary constraints")

        return [(k, s.report()) for k, s in self.summary_constraint_map.items()]

    def report(self) -> List[Tuple[str, Optional[List[tuple]]]]:
        l2 = [(k, s.report()) for k, s in self.summary_constraint_map.items()]
        return l2



import logging
from typing import Any, Dict, List

from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.projectors import SingleFieldProjector
from whylogs.core.proto import ColumnMessage
from whylogs.core.schema import ColumnSchema
from whylogs.core.view import ColumnProfileView

logger = logging.getLogger(__name__)


class ColumnProfile(object):
    def __init__(self, name: str, schema: ColumnSchema, cache_size: int):
        self._name = name
        self._schema = schema

        self._metrics = schema.get_metrics(name)
        self._projector = SingleFieldProjector(col_name=name)
        self._success_count = 0
        self._failure_count = 0
        logger.debug("Setting cache size for column: %s with size: %s", self._name, cache_size)
        self._cache_size = cache_size
        self._cache: List[Any] = []

    def track(self, row: Dict[str, Any]) -> None:
        value = self._projector.apply(row)
        if len(self._cache) < self._cache_size - 1:
            self._cache.append(value)
        else:
            self._cache.append(value)
            cache_list = self._cache
            self._cache = []
            self.track_column(cache_list)

    def flush(self) -> None:
        """Force emptying the cache and update the internal metrics."""

        logger.debug("Flushing out the cache for col: %s. Cache size: %s", self._name, self._cache_size)
        old_cache = self._cache
        self._cache = []
        self.track_column(old_cache)

    def track_column(self, series: Any) -> None:
        ex_col = PreprocessedColumn.apply(series)
        for metric in self._metrics.values():
            res = metric.columnar_update(ex_col)
            self._failure_count += res.failures
            self._success_count += res.successes

    def to_protobuf(self) -> ColumnMessage:
        self.flush()

        return self.view().to_protobuf()

    def view(self) -> ColumnProfileView:
        self.flush()

        return ColumnProfileView(
            metrics=self._metrics.copy(),
            success_count=self._success_count,
            failure_count=self._failure_count,
        )



from typing import Any, Dict


class SingleFieldProjector(object):
    def __init__(self, col_name: str) -> None:
        self._col_name = col_name

    def apply(self, row: Dict[str, Any]) -> Any:
        return row.get(self._col_name)



import logging
import tempfile
from datetime import datetime
from enum import Enum
from typing import Dict, List, Optional

from google.protobuf.message import DecodeError

from whylogs.core.configs import SummaryConfig
from whylogs.core.errors import DeserializationError
from whylogs.core.proto import (
    ChunkHeader,
    ChunkMessage,
    ChunkOffsets,
    ColumnMessage,
    DatasetProfileHeader,
    DatasetProperties,
    DatasetSegmentHeader,
    MetricComponentMessage,
)
from whylogs.core.stubs import pd
from whylogs.core.utils import read_delimited_protobuf, write_delimited_protobuf
from whylogs.core.utils.timestamp_calculations import to_utc_milliseconds
from whylogs.core.view.column_profile_view import ColumnProfileView

# Magic header for whylogs using the first 4 bytes
WHYLOGS_MAGIC_HEADER = "WHY1"
WHYLOGS_MAGIC_HEADER_LEN = 4

WHYLOGS_MAGIC_HEADER_BYTES = WHYLOGS_MAGIC_HEADER.encode("utf-8")

logger = logging.getLogger(__name__)


class SummaryType(str, Enum):
    COLUMN = "COLUMN"
    DATASET = "DATASET"


class DatasetProfileView(object):
    _columns: Dict[str, ColumnProfileView]

    def __init__(
        self, *, columns: Dict[str, ColumnProfileView], dataset_timestamp: datetime, creation_timestamp: datetime
    ):
        self._columns = columns.copy()
        self._dataset_timestamp = dataset_timestamp
        self._creation_timestamp = creation_timestamp

    @property
    def dataset_timestamp(self) -> datetime:
        return self._dataset_timestamp

    @property
    def creation_timestamp(self) -> datetime:
        return self._creation_timestamp

    def merge(self, other: "DatasetProfileView") -> "DatasetProfileView":
        all_names = set(self._columns.keys()).union(other._columns.keys())
        merged_columns: Dict[str, ColumnProfileView] = {}
        for n in all_names:
            lhs = self._columns.get(n)
            rhs = other._columns.get(n)

            res = lhs
            if lhs is None:
                res = rhs
            elif rhs is not None:
                res = lhs + rhs
            assert res is not None
            merged_columns[n] = res
        return DatasetProfileView(
            columns=merged_columns,
            dataset_timestamp=self._dataset_timestamp,
            creation_timestamp=self._creation_timestamp,
        )

    def get_column(self, col_name: str) -> Optional[ColumnProfileView]:
        return self._columns.get(col_name)

    def get_columns(self, col_names: Optional[List[str]] = None) -> Dict[str, ColumnProfileView]:
        if col_names:
            return {k: self._columns.get(k) for k in col_names}
        else:
            return {k: self._columns.get(k) for k in self._columns}

    def write(self, path: str) -> None:
        all_metric_component_names = set()

        # capture the list of all metric component paths
        for col in self._columns.values():
            all_metric_component_names.update(col.get_metric_component_paths())
        metric_name_list = list(all_metric_component_names)
        metric_name_list.sort()
        metric_name_indices: Dict[str, int] = {}
        metric_index_to_name: Dict[int, str] = {}
        for i in range(0, len(metric_name_list)):
            metric_name_indices[metric_name_list[i]] = i
            metric_index_to_name[i] = metric_name_list[i]

        column_chunk_offsets: Dict[str, ChunkOffsets] = {}
        with tempfile.TemporaryFile("w+b") as f:
            for col_name in sorted(self._columns.keys()):
                column_chunk_offsets[col_name] = ChunkOffsets(offsets=[f.tell()])

                col = self._columns[col_name]

                # for a given column, turn it into a ChunkMessage.
                indexed_component_messages: Dict[int, MetricComponentMessage] = {}
                metric_components = col.to_protobuf().metric_components
                for m_name, m_component in metric_components.items():
                    index = metric_name_indices.get(m_name)
                    if index is None:
                        raise ValueError(f"Missing metric from index mapping. Metric name: {m_name}")
                    indexed_component_messages[index] = m_component

                chunk_msg = ChunkMessage(metric_components=indexed_component_messages)
                chunk_header = ChunkHeader(type=ChunkHeader.ChunkType.COLUMN, length=chunk_msg.ByteSize())
                write_delimited_protobuf(f, chunk_header)
                f.write(chunk_msg.SerializeToString())

            total_len = f.tell()
            f.flush()

            properties = DatasetProperties(
                dataset_timestamp=to_utc_milliseconds(self._dataset_timestamp),
                creation_timestamp=to_utc_milliseconds(self._creation_timestamp),
            )
            dataset_header = DatasetProfileHeader(
                column_offsets=column_chunk_offsets,
                properties=properties,
                length=total_len,
                indexed_metric_paths=metric_index_to_name,
            )

            dataset_segment_header = DatasetSegmentHeader(
                has_segments=False,
            )

            with open(path, "w+b") as out_f:
                out_f.write(WHYLOGS_MAGIC_HEADER_BYTES)
                write_delimited_protobuf(out_f, dataset_segment_header)
                write_delimited_protobuf(out_f, dataset_header)

                f.seek(0)
                while f.tell() < total_len:
                    buffer = f.read(1024)
                    out_f.write(buffer)

    @classmethod
    def read(cls, path: str) -> "DatasetProfileView":
        with open(path, "r+b") as f:
            buf = f.read(WHYLOGS_MAGIC_HEADER_LEN)
            try:
                decoded_header = buf.decode("utf-8")
            except UnicodeDecodeError as e:
                raise DeserializationError("Invalid magic header. Decoder error: %s", e)

            if WHYLOGS_MAGIC_HEADER != decoded_header:
                raise DeserializationError(
                    f"Invalid magic header. Got: {decoded_header} but expecting: {WHYLOGS_MAGIC_HEADER}"
                )

            dataset_segment_header = read_delimited_protobuf(f, DatasetSegmentHeader)
            if dataset_segment_header.has_segments:
                raise DeserializationError("File contains segments. This is not supported yet")

            # TODO: handle segment data
            dataset_profile_header = read_delimited_protobuf(f, DatasetProfileHeader)
            if dataset_profile_header.ByteSize() == 0:
                raise DeserializationError("Missing valid dataset profile header")
            dataset_timestamp = datetime.fromtimestamp(dataset_profile_header.properties.dataset_timestamp / 1000.0)
            creation_timestamp = datetime.fromtimestamp(dataset_profile_header.properties.creation_timestamp / 1000.0)
            indexed_metric_paths = dataset_profile_header.indexed_metric_paths
            if len(indexed_metric_paths) < 1:
                logger.warning("Name index in the header is empty. Possible data corruption")

            start_offset = f.tell()

            columns = {}
            for col_name in sorted(dataset_profile_header.column_offsets.keys()):
                col_offsets = dataset_profile_header.column_offsets[col_name]
                all_metric_components: Dict[str, MetricComponentMessage] = {}
                for offset in col_offsets.offsets:
                    actual_offset = offset + start_offset
                    chunk_header = read_delimited_protobuf(f, proto_class_name=ChunkHeader, offset=actual_offset)
                    if chunk_header is None:
                        raise DeserializationError(
                            f"Missing or corrupt chunk header for column {col_name}. Offset: {actual_offset}"
                        )
                    if chunk_header.type != ChunkHeader.ChunkType.COLUMN:
                        raise DeserializationError(
                            f"Expecting chunk header type to be {ChunkHeader.ChunkType.COLUMN}, "
                            f"got {chunk_header.type}"
                        )

                    chunk_msg = ChunkMessage()
                    buf = f.read(chunk_header.length)
                    if len(buf) != chunk_header.length:
                        raise IOError(
                            f"Invalid message for {col_name}. Expecting buffer length of {chunk_header.length}, "
                            f"got {len(buf)}. "
                            f"Offset: {actual_offset}"
                        )
                    try:
                        chunk_msg.ParseFromString(buf)
                    except DecodeError:
                        raise DeserializationError(f"Failed to parse protobuf message for column: {col_name}")

                    for idx, metric_component in chunk_msg.metric_components.items():
                        full_name = indexed_metric_paths.get(idx)
                        if full_name is None:
                            raise ValueError(f"Missing metric name in the header. Index: {idx}")
                        all_metric_components[full_name] = metric_component

                column_msg = ColumnMessage(metric_components=all_metric_components)
                columns[col_name] = ColumnProfileView.from_protobuf(column_msg)
            return DatasetProfileView(
                columns=columns, dataset_timestamp=dataset_timestamp, creation_timestamp=creation_timestamp
            )

    def to_pandas(self, column_metric: Optional[str] = None, cfg: Optional[SummaryConfig] = None) -> pd.DataFrame:
        all_dicts = []
        for col_name, col in self._columns.items():
            sum_dict = col.to_summary_dict(column_metric=column_metric, cfg=cfg)
            sum_dict["column"] = col_name
            sum_dict["type"] = SummaryType.COLUMN
            all_dicts.append(sum_dict)
        df = pd.DataFrame(all_dicts)
        return df.set_index("column")



import unicodedata
from typing import Dict, List, Tuple

from typing_extensions import TypeAlias

from whylogs.core.metrics.compound_metric import CompoundMetric
from whylogs.core.metrics.metrics import DistributionMetric, OperationResult
from whylogs.core.preprocessing import PreprocessedColumn
from whylogs.core.proto import MetricMessage

ColumnSchema: TypeAlias = "ColumnSchema"  # type: ignore


DEFAULT_RANGES = {
    "emoji": (0x1F600, 0x1F64F),
    "control": (0x00, 0x1F),
    "digits": (0x30, 0x39),
    "latin-lower": (0x41, 0x5A),
    "latin-upper": (0x61, 0x7A),
    "basic-latin": (0x00, 0x7F),
    "extended-latin": (0x0080, 0x02AF),
}


_STRING_LENGTH = "string_length"


class UnicodeRangeMetric(CompoundMetric):
    """
    For string values, maintains a DistributionMetric for the counts of
    characters that fall within user-defined codepoint ranges.
    """

    range_definitions: Dict[str, Tuple[int, int]]

    def __init__(
        self,
        range_definitions: Dict[str, Tuple[int, int]] = None,
    ):
        """
        Parameters
        ----------
        range_definitions : Dict[str, Tuple[int, int]]
            Defines the character ranges to be counted. The key servers as
            the range name and should only contain alphanumeric, hyphen, and
            underscore characters. The tuple defines the Unicode codepoint
            range to be tracked. The string length is tracked under the key
            "STRING_LENGTH" so don't use that as a range name.
        """
        if range_definitions is None:
            range_definitions = DEFAULT_RANGES
        range_definitions["UNKNOWN"] = (0, 0)  # catchall for characters not in a defined range
        for key, range in range_definitions.items():
            if range[0] > range[1]:
                raise ValueError(f"Invalid codepoint range {key}")
            if range[0] < 0 or 0x10FFFF < range[1]:
                raise ValueError(f"Invalid codepoint range {key}")
            if ":" in key or "/" in key:
                raise ValueError(f"Invalid range name {key}")
        if _STRING_LENGTH in range_definitions:
            raise ValueError("STRING_LENGTH cannot be used as a range name")

        self.range_definitions = range_definitions
        from whylogs.core.schema import ColumnSchema

        submetrics = {key: DistributionMetric.zero(ColumnSchema(dtype=int)) for key in range_definitions.keys()}
        submetrics[_STRING_LENGTH] = DistributionMetric.zero(ColumnSchema(dtype=int))
        super(type(self), self).__init__(submetrics)  # type: ignore

    @property
    def namespace(self) -> str:
        return "unicode_range"

    def merge(self, other: "UnicodeRangeMetric") -> "UnicodeRangeMetric":
        submetrics = self.merge_submetrics(other)
        result = UnicodeRangeMetric(self.range_definitions)
        result.submetrics = submetrics
        return result

    def columnar_update(self, view: PreprocessedColumn) -> OperationResult:
        data = (
            view.pandas.strings.to_list() if view.pandas.strings is not None and not view.pandas.strings.empty else []
        )
        data = (data + view.list.strings) if view.list.strings else data
        range_data: Dict[str, List[int]] = {range_name: [] for range_name in self.range_definitions.keys()}
        lengths: List[int] = []
        for value in data:
            lengths.append(len(value))
            range_counter: Dict[str, int] = {range_name: 0 for range_name in self.range_definitions.keys()}
            # TODO: need to transform to utf-32 or handle surrogates
            for char in unicodedata.normalize("NFD", value).lower():
                found = False
                for range_name, range_limits in self.range_definitions.items():
                    if range_limits[0] <= ord(char) <= range_limits[1]:
                        range_counter[range_name] += 1
                        found = True
                    if not found:
                        range_counter["UNKNOWN"] += 1

            for range_name, range_count in range_counter.items():
                range_data[range_name].append(range_count)

        submetric_col = PreprocessedColumn()
        submetric_col.list.ints = lengths
        self.submetrics[_STRING_LENGTH].columnar_update(submetric_col)
        for range_name, range_list in range_data.items():
            submetric_col.list.ints = range_list
            self.submetrics[range_name].columnar_update(submetric_col)
        return OperationResult.ok(len(data))

    @classmethod
    def zero(cls, schema: ColumnSchema) -> "UnicodeRangeMetric":
        return cls(schema.cfg.unicode_ranges)

    @classmethod
    def from_protobuf(cls, msg: MetricMessage) -> "UnicodeRangeMetric":
        submetrics = cls.submetrics_from_protobuf(msg)
        # The MetricMessage doesn't contain the range definitions, so we preserve
        # the range names with empty bounds
        ranges = {sub_name: (0, 0) for sub_name in submetrics.keys()}
        del ranges[_STRING_LENGTH]
        result = cls(ranges)
        result.submetrics = submetrics
        return result



import logging
from dataclasses import dataclass
from typing import Any, Dict, Mapping, Optional, TypeVar

from whylogs.core.datatypes import StandardTypeMapper, TypeMapper
from whylogs.core.metrics import Metric
from whylogs.core.resolvers import Resolver, StandardResolver
from whylogs.core.stubs import pd

logger = logging.getLogger(__name__)

LARGE_CACHE_SIZE_LIMIT = 1024 * 100

T = TypeVar("T", bound="DatasetSchema")


@dataclass(frozen=True)
class ColumnConfig:
    hll_lg_k: int = 12
    kll_k: int = 256
    fi_lg_max_k: int = 10  # 128 entries
    fi_disabled: bool = False
    unicode_ranges = {
        "emoji": (0x1F600, 0x1F64F),
        "control": (0x00, 0x1F),
        "digits": (0x30, 0x39),
        "latin-lower": (0x41, 0x5A),
        "latin-upper": (0x61, 0x7A),
        "basic-latin": (0x00, 0x7F),
        "extended-latin": (0x0080, 0x02AF),
    }


class DatasetSchema:
    """
    Defines the schema for tracking metrics in whylogs.

    In order to customize your tracking, you can extend this class to specify your own column schema
    or your own type resolution. Otherwise, you can just use the default DatasetSchema object.

    Schema objects are also used to group datasets together.

    Attributes:
        types: required. a dictionay of column name to the Python type.

        default_configs: optional. Options to configure various behavior of whylogs.

        type_mapper: Optional. a mapper that transates the Python type to standardized whylogs :class:`DataType` object.

        resolvers: Optional. an object that defines how to map from a column name, a whylogs :class:`DataType` and a
                   schema to metrics.


    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from whylogs.core import DatasetSchema, DatasetProfile
    >>>
    >>> class MySchema(DatasetSchema):
    ...    types = {
    ...        "col1": str,
    ...        "col2": np.int32,
    ...        "col3": pd.Categorical(categories=('foo', 'bar'), ordered=True)
    ...    }
    ...    cache_size = 12
    >>> schema = MySchema()
    >>> prof = DatasetProfile(MySchema())
    >>> df = pd.DataFrame({"col1": ['foo'], "col2": np.array([1], dtype=np.int32), "col3": ['bar']})
    >>> prof.track(pandas=df)

    """

    types: Dict[str, Any] = {}
    default_configs: ColumnConfig = ColumnConfig()
    type_mapper: TypeMapper = StandardTypeMapper()
    resolvers: Resolver = StandardResolver()
    cache_size: int = 1024
    schema_based_automerge: bool = False

    def __init__(self) -> None:
        self._columns = {}

        if self.cache_size < 0:
            logger.warning("Negative cache size value. Disabling caching")
            self.cache_size = 0

        if self.cache_size > LARGE_CACHE_SIZE_LIMIT:
            logger.warning(
                "Cache size value %s is larger than recommended limit of %s",
                self.cache_size,
                LARGE_CACHE_SIZE_LIMIT,
            )

        if self.types:
            for col, tpe in self.types.items():
                self._columns[col] = ColumnSchema(
                    dtype=tpe, resolver=self.resolvers, type_mapper=self.type_mapper, cfg=self.default_configs
                )

    def copy(self) -> "DatasetSchema":
        """Returns a new instance of the same underlying schema"""
        copy = DatasetSchema()
        copy._columns = self._columns.copy()
        copy.cache_size = self.cache_size
        copy.type_mapper = self.type_mapper
        copy.resolvers = self.resolvers

        return copy

    def resolve(self, *, pandas: Optional[pd.DataFrame] = None, row: Optional[Mapping[str, Any]] = None) -> bool:
        if pandas is not None:
            return self._resolve_pdf(pandas)

        if row is not None:
            for k, v in row.items():
                if k in self._columns:
                    continue

                self._columns[k] = ColumnSchema(
                    dtype=type(v),
                    resolver=self.resolvers,
                    type_mapper=self.type_mapper,
                )
            return True
        raise NotImplementedError

    def _resolve_pdf(self, df: pd.DataFrame, force_resolve: bool = False) -> bool:
        """
        Resolve ColumnSchema from the dataframe. We only resolve newly detected
        columns unless `force_resolve` is set to True.
        """
        col_names = df.dtypes.keys()
        dirty = False
        for col_name in col_names:
            if not force_resolve and col_name in self._columns:
                logger.debug("column is already configured")
                continue

            col_dtype = df.dtypes[col_name]
            self._columns[col_name] = ColumnSchema(
                dtype=col_dtype,
                resolver=self.resolvers,
                type_mapper=self.type_mapper,
            )
            dirty = True

        return dirty

    def get_col_names(self) -> frozenset:
        return frozenset(self._columns.keys())

    def get(self, name: str) -> Optional["ColumnSchema"]:
        return self._columns.get(name)


@dataclass(frozen=True)
class ColumnSchema:
    """
    Schema of a column.

    The main goal is to specify the data type.
    On top of that, users can configure their own tracker resolution logic (mapping a type to a list of tracker
    factories) and any additional trackers here.
    """

    dtype: Any
    cfg: ColumnConfig = ColumnConfig()
    type_mapper: TypeMapper = StandardTypeMapper()
    resolver: Resolver = StandardResolver()

    def get_metrics(self, name: str) -> Dict[str, Metric]:
        return self.resolver.resolve(name=name, why_type=self.type_mapper(self.dtype), column_schema=self)



import datetime
import logging
import os
import tempfile
from typing import Any, Optional

import requests  # type: ignore
import whylabs_client
from whylabs_client.api.log_api import LogApi
from whylabs_client.model.log_async_request import LogAsyncRequest
from whylabs_client.rest import ForbiddenException

from whylogs.api.writer import Writer
from whylogs.core import DatasetProfileView
from whylogs.core.errors import BadConfigError

FIVE_MINUTES_IN_SECONDS = 60 * 5
logger = logging.getLogger(__name__)


class WhyLabsWriter(Writer):
    """
    A WhyLogs writer to upload DatasetProfileView's onto the WhyLabs platform.

    >**IMPORTANT**: In order to correctly send your profiles over, make sure you have
    the following environment variables set: `[WHYLABS_ORG_ID, WHYLABS_API_KEY]`. You
    can also set them with the option method or within the constructor, although it
    is highly recommended you don't persist credentials in code!

    Parameters
    ----------
    org_id: str, optional
        Your organization ID on the WhyLabs platform. If not informed, will get the mentioned environment
        variable instead.

    api_key: str, optional
        The API Key to authenticate with WhyLabs. If not informed, will get the mentioned environment
        variable instead.

    dataset_id: str, optional
        The Dataset ID, often referred to as model ID, that will tie the profile to your tracked data.
        If not informed, will get the mentioned environment variable instead.
    Returns
    -------
        None

    Examples
    --------
    An example usage of this method can be represented with the simple code above. Here we
    are assuming the user has already assigned a pandas.DataFrame to the `df` variable.

    ```python
    import whylogs as why

    profile = why.log(pandas=df)
    profile.writer("whylabs").write()
    ```

    """

    def __init__(self, org_id: Optional[str] = None, api_key: Optional[str] = None, dataset_id: Optional[str] = None):
        self._org_id = org_id or os.environ.get("WHYLABS_ORG_ID")
        self._api_key = api_key or os.environ.get("WHYLABS_API_KEY")
        self._dataset_id = dataset_id or os.environ.get("WHYLABS_DEFAULT_DATASET_ID")
        self.whylabs_api_endpoint = os.environ.get("WHYLABS_API_ENDPOINT") or "https://api.whylabsapp.com"
        self._whylabs_v1_enabled = os.environ.get("WHYLABS_V1_ENABLED")

    def check_interval(self, interval_seconds: int):
        if interval_seconds < FIVE_MINUTES_IN_SECONDS:
            raise BadConfigError("Bad WhyLabsWriter config: interval must be greater or equal to five minutes")

    def option(
        self, org_id: Optional[str] = None, dataset_id: Optional[str] = None, api_key: Optional[str] = None
    ) -> None:
        if dataset_id is not None:
            self._dataset_id = dataset_id
        if org_id is not None:
            self._org_id = org_id
        if api_key is not None:
            self._api_key = api_key

    def write(self, profile: DatasetProfileView, dataset_id: Optional[str] = None) -> Any:
        if self._whylabs_v1_enabled != "True":
            raise ValueError("The Whylabs writer is currently not supported!")

        if dataset_id is not None:
            self._dataset_id = dataset_id

        with tempfile.NamedTemporaryFile() as tmp_file:
            profile.write(path=tmp_file.name)
            tmp_file.flush()

            dataset_timestamp = profile.dataset_timestamp or datetime.datetime.now(datetime.timezone.utc)
            dataset_timestamp = int(dataset_timestamp.timestamp() * 1000)
            return self._upload_whylabs(dataset_timestamp=dataset_timestamp, profile_path=tmp_file.name)

    def _upload_whylabs(
        self, dataset_timestamp: int, profile_path: str, upload_url: Optional[str] = None
    ) -> requests.Response:
        if self._org_id is None:
            raise EnvironmentError(
                "Missing organization ID. Specify it via option or WHYLABS_DEFAULT_ORG_ID " "environment variable"
            )
        if self._dataset_id is None:
            raise EnvironmentError(
                "Missing dataset ID. Specify it via WHYLABS_DEFAULT_DATASET_ID environment "
                "variable or on your write method"
            )

        upload_url = upload_url or self._get_upload_url(dataset_timestamp=dataset_timestamp)

        try:
            with open(profile_path, "rb") as f:
                http_response = requests.put(upload_url, data=f.read())
                if http_response.status_code == 200:
                    logger.info(
                        f"Done uploading {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                        f"{self.whylabs_api_endpoint} with API token ID: {self._api_key}"
                    )
                return http_response
        except requests.RequestException as e:
            logger.info(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to "
                + f"{self.whylabs_api_endpoint}. Error occurred: {e}"
            )

    def _get_or_create_api_log_client(self) -> LogApi:
        environment_api_key = os.environ["WHYLABS_API_KEY"]
        _api_log_client = None

        if environment_api_key is not None and self._api_key != environment_api_key:
            updated_key = os.environ["WHYLABS_API_KEY"]
            logger.warning(f"Updating API key ID from: {self._api_key} to: {updated_key}")
            self._api_key = updated_key
            config = whylabs_client.Configuration(
                host=self.whylabs_api_endpoint, api_key={"ApiKeyAuth": self._api_key}, discard_unknown_keys=True
            )
            _api_log_client = whylabs_client.ApiClient(config)

        if _api_log_client is None:
            config = whylabs_client.Configuration(
                host=self.whylabs_api_endpoint, api_key={"ApiKeyAuth": self._api_key}, discard_unknown_keys=True
            )
            _api_log_client = whylabs_client.ApiClient(config)
        return LogApi(_api_log_client)

    @staticmethod
    def _build_log_async_request(dataset_timestamp):
        request = LogAsyncRequest(dataset_timestamp=dataset_timestamp, segment_tags=[])
        return request

    def _post_log_async(self, request, dataset_timestamp):
        log_api = self._get_or_create_api_log_client()
        try:
            result = log_api.log_async(org_id=self._org_id, dataset_id=self._dataset_id, log_async_request=request)
            return result
        except ForbiddenException as e:
            logger.exception(
                f"Failed to upload {self._org_id}/{self._dataset_id}/{dataset_timestamp} to {self.whylabs_api_endpoint} with API token ID: {self._api_key[:10]}"
            )
            raise e

    def _get_upload_url(self, dataset_timestamp: int):
        request = self._build_log_async_request(dataset_timestamp)
        log_api = self._post_log_async(request=request, dataset_timestamp=dataset_timestamp)
        upload_url = log_api["upload_url"]
        return upload_url



import os
from typing import Optional

from whylogs.api.writer import Writer
from whylogs.core import DatasetProfileView


class LocalWriter(Writer):
    def __init__(self, base_dir: Optional[str] = None, base_name: Optional[str] = None) -> None:
        if base_dir is None:
            base_dir = os.getcwd()
        if base_name is None:
            base_name = "profile"

        self._base_dir = base_dir
        self._base_name = base_name

    def write(self, profile: DatasetProfileView, dest: Optional[str] = None) -> None:
        if dest is None:
            dest = f"{self._base_name}_{profile.creation_timestamp}.bin"
        full_path = os.path.join(self._base_dir, dest)
        profile.write(full_path)

    def option(self, base_dir: Optional[str] = None, base_name: Optional[str] = None) -> "LocalWriter":  # type: ignore
        if base_dir is not None:
            self._base_dir = base_dir
        if base_name is not None:
            self._base_name = base_name

        return self



import logging
import os.path
import time
from datetime import datetime, timezone
from typing import Any, Dict, Mapping, Optional

from .column_profile import ColumnProfile
from .schema import DatasetSchema
from .stubs import pd
from .view import DatasetProfileView

logger = logging.getLogger(__name__)

_LARGE_CACHE_SIZE_LIMIT = 1024 * 100


class DatasetProfile(object):
    """
    Dataset profile represents a collection of in-memory profiling stats for a dataset.

    Args:
        schema: :class:`DatasetSchema`, optional
            An object that represents the data column names and types
        dataset_timestamp: int, optional
            A timestamp integer that best represents the date tied to the dataset generation.
            i.e.: A January 1st 2019 Sales Dataset will have 1546300800000 as the timestamp in miliseconds (UTC).
            If None is provided, it will take the current timestamp as default
        creation_timestamp: int, optional
            The timestamp tied to the exact moment when the :class:`DatasetProfile` is created.
            If None is provided, it will take the current timestamp as default
    """

    def __init__(
        self,
        schema: Optional[DatasetSchema] = None,
        dataset_timestamp: Optional[datetime] = None,
        creation_timestamp: Optional[datetime] = None,
    ):
        if schema is None:
            schema = DatasetSchema()
        now = datetime.utcnow()
        self._dataset_timestamp = dataset_timestamp or now
        self._creation_timestamp = creation_timestamp or now
        self._schema = schema
        self._columns: Dict[str, ColumnProfile] = dict()
        self._is_active = False
        self._track_count = 0

    @property
    def creation_timestamp(self) -> datetime:
        return self._creation_timestamp

    @property
    def dataset_timestamp(self) -> datetime:
        return self._dataset_timestamp

    @property
    def is_active(self) -> bool:
        """Returns True if the profile tracking code is currently running."""
        return self._is_active

    @property
    def is_empty(self) -> bool:
        """Returns True if the profile tracking code is currently running."""
        return self._track_count == 0

    def set_dataset_timestamp(self, dataset_timestamp: datetime) -> None:
        if dataset_timestamp.tzinfo is None:
            logger.warning("No timezone set in the datetime_timestamp object. Default to local timezone")

        self._dataset_timestamp = dataset_timestamp.astimezone(tz=timezone.utc)

    def track(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Mapping[str, Any]] = None,
    ) -> None:
        try:
            self._is_active = True
            self._track_count += 1
            self._do_track(obj, pandas=pandas, row=row)
        finally:
            self._is_active = False

    def _do_track(
        self,
        obj: Any = None,
        *,
        pandas: Optional[pd.DataFrame] = None,
        row: Optional[Mapping[str, Any]] = None,
    ) -> None:
        if obj is not None:
            if pandas is not None:
                raise ValueError("Cannot pass both obj and pandas params")
            if row is not None:
                raise ValueError("Cannot pass both obj and row params")

            if isinstance(obj, pd.DataFrame):
                pandas = obj
            elif isinstance(obj, (dict, Dict, Mapping)):
                row = obj

        if pandas is not None and row is not None:
            raise ValueError("Cannot pass both pandas and row params")

        # TODO: do this less frequently when operating at row level
        dirty = self._schema.resolve(pandas=pandas, row=row)
        if dirty:
            new_cols = self._schema.get_col_names().difference(self._columns.keys())
            for col in new_cols:
                col_schema = self._schema.get(col)
                if col_schema:
                    self._columns[col] = ColumnProfile(name=col, schema=col_schema, cache_size=self._schema.cache_size)
                else:
                    logger.warning("Encountered a column without schema: %s", col)

        if pandas is not None:
            for k in pandas.keys():
                self._columns[k].track_column(pandas[k])
            return

        if row is not None:
            for k in row.keys():
                self._columns[k].track_column([row[k]])
            return

        raise NotImplementedError

    def view(self) -> DatasetProfileView:
        columns = {}
        for c_name, c in self._columns.items():
            columns[c_name] = c.view()
        return DatasetProfileView(
            columns=columns, dataset_timestamp=self.dataset_timestamp, creation_timestamp=self.creation_timestamp
        )

    def flush(self) -> None:
        for col in self._columns.values():
            col.flush()

    def write(self, path_or_base_dir: str) -> None:
        if not path_or_base_dir.endswith(".bin"):
            output_path = os.path.join(path_or_base_dir, f"profile.{int(round(time.time() * 1000))}.bin")
        else:
            output_path = path_or_base_dir

        self.view().write(output_path)
        logger.debug("Wrote profile to path: %s", output_path)

    @classmethod
    def read(cls, input_path: str) -> DatasetProfileView:
        return DatasetProfileView.read(input_path)

    def __repr__(self) -> str:
        return f"DatasetProfile({len(self._columns)} columns). Schema: {str(self._schema)}"



import datetime
from typing import Dict

import whylogs_sketching as ds  # type: ignore

from whylogs.core import ColumnProfileView, DatasetProfileView
from whylogs.core.errors import DeserializationError
from whylogs.core.metrics import (
    ColumnCountsMetric,
    DistributionMetric,
    FrequentItemsMetric,
    IntsMetric,
    StandardMetric,
    TypeCountersMetric,
)
from whylogs.core.metrics.metric_components import (
    FractionalComponent,
    FrequentStringsComponent,
    IntegralComponent,
    KllComponent,
    MaxIntegralComponent,
    MinIntegralComponent,
)
from whylogs.core.proto.v0 import ColumnMessageV0, DatasetProfileMessageV0, InferredType
from whylogs.core.utils import read_delimited_protobuf


def read_v0_to_view(path: str) -> DatasetProfileView:
    with open(path, "r+b") as f:
        v0_msg = read_delimited_protobuf(f, DatasetProfileMessageV0)
        if v0_msg is None:
            raise DeserializationError("Unexpected empty message")
        return v0_to_v1_view(v0_msg)


def v0_to_v1_view(msg: DatasetProfileMessageV0) -> DatasetProfileView:
    columns: Dict[str, ColumnProfileView] = {}

    dataset_timestamp = datetime.datetime.fromtimestamp(msg.properties.data_timestamp / 1000.0)
    creation_timestamp = datetime.datetime.fromtimestamp(msg.properties.session_timestamp / 1000.0)

    for col_name, col_msg in msg.columns.items():
        dist_metric = _extract_dist_metric(col_msg)
        fs = FrequentStringsComponent(ds.frequent_strings_sketch.deserialize(col_msg.frequent_items.sketch))
        fi_metric = FrequentItemsMetric(frequent_strings=fs)
        count_metrics = _extract_col_counts(col_msg)
        type_counters_metric = _extract_type_counts_metric(col_msg)
        int_metric = _extract_ints_metric(col_msg)

        columns[col_name] = ColumnProfileView(
            metrics={
                StandardMetric.distribution.name: dist_metric,
                StandardMetric.frequent_items.name: fi_metric,
                StandardMetric.counts.name: count_metrics,
                StandardMetric.types.name: type_counters_metric,
                StandardMetric.cardinality.name: type_counters_metric,
                StandardMetric.ints.name: int_metric,
            }
        )

    return DatasetProfileView(
        columns=columns, dataset_timestamp=dataset_timestamp, creation_timestamp=creation_timestamp
    )


def _extract_ints_metric(msg: ColumnMessageV0) -> IntsMetric:
    int_max = msg.numbers.longs.max
    int_min = msg.numbers.longs.min
    return IntsMetric(max=MaxIntegralComponent(int_max), min=MinIntegralComponent(int_min))


def _extract_type_counts_metric(msg: ColumnMessageV0) -> TypeCountersMetric:
    int_count = msg.schema.typeCounts.get(InferredType.INTEGRAL)
    bool_count = msg.schema.typeCounts.get(InferredType.BOOLEAN)
    frac_count = msg.schema.typeCounts.get(InferredType.FRACTIONAL)
    string_count = msg.schema.typeCounts.get(InferredType.STRING)
    obj_count = msg.schema.typeCounts.get(InferredType.UNKNOWN)
    return TypeCountersMetric(
        integral=IntegralComponent(int_count or 0),
        fractional=IntegralComponent(frac_count or 0),
        boolean=IntegralComponent(bool_count or 0),
        string=IntegralComponent(string_count or 0),
        object=IntegralComponent(obj_count or 0),
    )


def _extract_col_counts(msg: ColumnMessageV0) -> ColumnCountsMetric:
    count_n = msg.counters.count
    count_null = msg.counters.null_count
    return ColumnCountsMetric(n=IntegralComponent(count_n or 0), null=IntegralComponent(count_null.value or 0))


def _extract_dist_metric(msg: ColumnMessageV0) -> DistributionMetric:
    floats_sk = ds.kll_floats_sketch.deserialize(msg.numbers.histogram)
    doubles_sk: ds.kll_doubles_sketch = ds.kll_floats_sketch.float_to_doubles(floats_sk)
    dist_mean = msg.numbers.variance.mean
    dist_m2 = msg.numbers.variance.sum
    return DistributionMetric(
        kll=KllComponent(doubles_sk),
        mean=FractionalComponent(dist_mean),
        m2=FractionalComponent(dist_m2),
    )



from typing import Optional

from whylogs.api.writer import Writer
from whylogs.core import DatasetProfileView


class MlflowWriter(Writer):
    def write(self, profile: DatasetProfileView, dest: Optional[str] = None) -> None:
        pass

    def option(self, param: Optional[str] = None) -> "MlflowWriter":  # type: ignore
        pass



from abc import ABC, abstractmethod
from typing import Any, Optional, TypeVar

from whylogs.core import DatasetProfileView

T = TypeVar("T", bound="Writer")


class Writer(ABC):

    # noinspection PyMethodMayBeStatic
    def check_interval(self, interval_seconds: int) -> None:
        """Validate an interval configuration for a given writer.

        Some writer only accepts certain interval configuration. By default, this should return True for a valid
        non-negative interval."""

    @abstractmethod
    def write(self, profile: DatasetProfileView, dest: Optional[str] = None) -> None:
        pass

    @abstractmethod
    def option(self: T, **kwargs: Any) -> T:
        pass



from whylogs.core.constraints.constraints import (
    DatasetConstraints,
    SummaryConstraint,
    SummaryConstraints,
)

__ALL__ = [
    # column
    SummaryConstraint,
    SummaryConstraints,
    DatasetConstraints,
]



from .protobuf_utils import read_delimited_protobuf, write_delimited_protobuf
from .stats_calculations import get_distribution_metrics

__ALL__ = [
    read_delimited_protobuf,
    write_delimited_protobuf,
    get_distribution_metrics,
]



import atexit
import hashlib
import http.client
import json
import logging
import os
import site
import socket
import sys
import uuid
from datetime import datetime
from threading import Thread
from typing import Any, Dict
from urllib import request

import whylogs

HEAP_APPID_PROD = "2170415941"

HEAP_ENDPOINT = "https://heapanalytics.com/api/track"
HEAP_HEADERS = {"Content-Type": "application/json"}
TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%fZ"
logger = logging.getLogger(__name__)

ANALYTICS_OPT_OUT = "WHYLOGS_NO_ANALYTICS"

# Flag to disable it internally
_TELEMETRY_DISABLED = False


_SITE_PACKAGES = site.getsitepackages()


def emit_usage(event: str) -> None:
    if _TELEMETRY_DISABLED:
        return

    t = Thread(target=_do_emit_usage, args=(event,))
    t.start()

    atexit.register(t.join)


_metadata = None
_identity = None


def _do_emit_usage(event: str) -> None:
    if os.getenv(ANALYTICS_OPT_OUT) is not None:
        logger.debug("Opted out of usage statistics. Skipping.")
        return

    logger.debug("Telemetry opted in. Emitting usage statistics")

    global _identity
    global _metadata
    if _identity is None:
        _identity = _calc_identity()
    if _metadata is None:
        _metadata = _build_metadata()

    _send_heap_event(event, _identity, _metadata)


def _calc_identity() -> str:
    try:
        hashed_computer_name = hashlib.sha512(bytes(socket.gethostname(), encoding="utf8"))
        return hashed_computer_name.hexdigest()
    except socket.timeout as exc:
        logger.debug(
            "Socket timeout when trying to get the computer name. Exception: %s",
            exc,
        )
        return uuid.uuid4().hex


def _build_metadata() -> Dict[str, Any]:
    """Hash system and project data to send to Heap."""

    project_version = whylogs.__version__
    (major, minor, macro, _, _) = sys.version_info

    metadata = {
        "project_version": project_version,
        "python_version": f"{major}.{minor}.{macro}",
        "python_version_full": sys.version,
        "terminal": _get_terminal_mode(),
        "os": sys.platform,
        "conda": ("CONDA_DEFAULT_ENV" in os.environ),
        "venv": ("VIRTUAL_ENV" in os.environ),
        "environment": _get_environment(),
    }

    # track various integrations
    integrations = {
        "numpy": _has_lib("numpy"),
        "pandas": _has_lib("pandas"),
        "mlflow": _has_lib("mlflow"),
        "dask": _has_lib("dask"),
        "ray": _has_lib("ray"),
        "airflow": _has_lib("airflow"),
        "pyspark": _has_lib("pyspark"),
        "flyte": _has_lib("flyte"),
        "kafka": _has_lib("kafka"),
    }
    for k in list(integrations.keys()):
        if integrations.get(k) is False:
            integrations.pop(k)

    # add integration metadata
    metadata.update(integrations)
    return metadata


def _get_heap_app_id() -> str:
    """
    Get the Heap App ID to send the data to.
    This will be the development ID if it's set as an
    environment variable, otherwise it will be the production ID.
    """
    return os.environ.get("HEAP_APPID_DEV", HEAP_APPID_PROD)


def _send_heap_event(event_name: str, identity: str, properties: Dict[str, Any] = None) -> None:
    data = {
        "app_id": _get_heap_app_id(),
        "identity": identity,
        "event": event_name,
        "timestamp": datetime.utcnow().strftime(TIMESTAMP_FORMAT),
        "properties": properties or {},
    }
    global _TELEMETRY_DISABLED
    json_data = json.dumps(data).encode()
    req = request.Request(HEAP_ENDPOINT, data=json_data, method="POST")
    req.add_header("Content-Type", "application/json")
    resp: http.client.HTTPResponse = None  # type: ignore
    try:
        resp = request.urlopen(req, timeout=3)
        if resp.status != 200:
            logger.warning("Unable to send usage stats. Disabling stats collection.")
            _TELEMETRY_DISABLED = True
        logger.debug("Response: %s", resp.read())
    except:  # noqa
        logger.warning("Connection error. Skip stats collection.")
        _TELEMETRY_DISABLED = True

    finally:
        if resp is not None:
            resp.close()


def _get_terminal_mode() -> str:
    try:
        from IPython.core.getipython import get_ipython  # type: ignore

        ipython = get_ipython()
        if ipython is not None:
            return ipython.__class__.__name__
    except:  # noqa
        pass

    if hasattr(sys, "ps1"):
        return "shell"
    else:
        return "headless"


def _get_environment() -> str:
    environments_dict = {
        "GITHUB_ACTION": "github_action",
        "GITLAB_CI": "gitlab_ci",
        "BINDER_PORT": "binder",
        "PYCHARM_HOSTED": "pycharm",
        "SM_CURRENT_HOST": "sagemaker",
        "DATABRICKS_RUNTIME_VERSION": "databricks",
        "COLAB_GPU": "colab",
        "KAGGLE_KERNEL_RUN_TYPE": "kaggle",
        "DEEPNOTE_PROJECT_ID": "deepnote",
    }

    for key, value in environments_dict.items():
        if key in os.environ:
            return value
    return "unknown"


def _has_lib(lib_name: str) -> bool:
    try:
        for p in _SITE_PACKAGES:
            if os.path.exists(os.path.join(p, lib_name)):
                return True
    except:  # noqa
        pass

    return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['whylogs',
 'whylogs.api',
 'whylogs.api.logger',
 'whylogs.api.pyspark.experimental',
 'whylogs.api.usage_stats',
 'whylogs.api.writer',
 'whylogs.core',
 'whylogs.core.constraints',
 'whylogs.core.metrics',
 'whylogs.core.proto',
 'whylogs.core.proto.v0',
 'whylogs.core.utils',
 'whylogs.core.view',
 'whylogs.migration',
 'whylogs.viz',
 'whylogs.viz.utils']

package_data = \
{'': ['*'],
 'whylogs.viz': ['html/*',
                 'html/css/*',
                 'html/fonts/*',
                 'html/images/*',
                 'html/js/*',
                 'html/templates/*']}

install_requires = \
['protobuf>=3.15.5', 'typing-extensions', 'whylogs-sketching==3.4.0.dev9']

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=0.23'],
 'docs': ['sphinx',
          'sphinx-autoapi',
          'sphinx-copybutton>=0.5.0,<0.6.0',
          'myst-parser[sphinx]>=0.17.2,<0.18.0',
          'furo>=2022.3.4,<2023.0.0',
          'sphinx-autobuild>=2021.3.14,<2022.0.0',
          'sphinxext-opengraph>=0.6.3,<0.7.0',
          'sphinx-inline-tabs',
          'nbsphinx>=0.8.8,<0.9.0'],
 's3': ['boto3>=1.22.13,<2.0.0'],
 'spark': ['pyarrow>=8.0.0,<9.0.0', 'pyspark>=3.0.0,<4.0.0'],
 'viz': ['pybars3>=0.9,<0.10', 'ipython', 'scipy<1.8'],
 'whylabs': ['whylabs-client>=0.3.0,<0.4.0']}

setup_kwargs = {
    'name': 'whylogs',
    'version': '1.0.0rc1',
    'description': 'Profile and monitor your ML data pipeline end-to-end',
    'long_description': '<img src="https://static.scarf.sh/a.png?x-pxid=bc3c57b0-9a65-49fe-b8ea-f711c4d35b82" /><p align="center">\n<img src="https://i.imgur.com/nv33goV.png" width="35%"/>\n</br>\n\n<h1 align="center">The open source standard for data logging\n </h1>\n  <h3 align="center">\n   <a href="https://docs.whylabs.ai/docs/"><b>Documentation</b></a> &bull;\n   <a href="https://bit.ly/whylogsslack"><b>Slack Community</b></a> &bull;\n   <a href="https://github.com/whylabs/whylogs#python-quickstart"><b>Python Quickstart</b></a>\n </h3>\n\n<p align="center">\n<a href="https://github.com/whylabs/whylogs-python/blob/mainline/LICENSE" target="_blank">\n    <img src="http://img.shields.io/:license-Apache%202-blue.svg" alt="License">\n</a>\n<a href="https://badge.fury.io/py/whylogs" target="_blank">\n    <img src="https://badge.fury.io/py/whylogs.svg" alt="PyPi Version">\n</a>\n<a href="https://github.com/python/black" target="_blank">\n    <img src="https://img.shields.io/badge/code%20style-black-000000.svg" alt="Code style: black">\n</a>\n<a href="https://pepy.tech/project/whylogs" target="_blank">\n    <img src="https://pepy.tech/badge/whylogs" alt="PyPi Downloads">\n</a>\n<a href="bit.ly/whylogs" target="_blank">\n    <img src="https://github.com/whylabs/whylogs-python/workflows/whylogs%20CI/badge.svg" alt="CI">\n</a>\n<a href="https://codeclimate.com/github/whylabs/whylogs-python/maintainability" target="_blank">\n    <img src="https://api.codeclimate.com/v1/badges/442f6ca3dca1e583a488/maintainability" alt="Maintainability">\n</a>\n</p>\n\n## What is whylogs\n\nwhylogs is an open source library for logging any kind of data. With whylogs, users are able to generate summaries of\ntheir datasets (called _whylogs profiles_) which they can use to:\n\n1. Track changes in their dataset\n2. Create _data constraints_ to know whether their data looks they way it should\n3. Quickly visualize key summary statistics about their datasets\n\nThese three functionalities enable a variety of use cases for data scientists, machine learning engineers, and data\nengineers:\n\n- Detecting data drift (and resultant ML model performance degradation)\n- Data quality validation\n- Exploratory data analysis via data profiling\n- Tracking data for ML experiments\n- And many more\n\nwhylogs can be run in Python or [Apache Spark](https://docs.whylabs.ai/docs/spark-integration) (both PySpark and Scala)\nenvironments on a variety of [data types](#data-types). We [integrate](#integrations) with lots of other tools including\nPandas, [AWS Sagemaker](https://aws.amazon.com/blogs/startups/preventing-amazon-sagemaker-model-degradation-with-whylabs/)\n, [MLflow](https://docs.whylabs.ai/docs/mlflow-integration)\n, [Flask](https://whylabs.ai/blog/posts/deploy-and-monitor-your-ml-application-with-flask-and-whylabs)\n, [Ray](https://docs.whylabs.ai/docs/ray-integration)\n, [RAPIDS](https://whylabs.ai/blog/posts/monitoring-high-performance-machine-learning-models-with-rapids-and-whylogs)\n, [Apache Kafka](https://docs.whylabs.ai/docs/kafka-integration), and more.\n\nIf you have any questions, comments, or just want to hang out with us, please\njoin [our Slack Community](http://join.slack.whylabs.ai/). In addition to joining the Slack Community, you can also help\nthis project by giving us a ⭐ in the upper right corner of this page.\n\n## Python Quickstart<a name="python-quickstart" />\n\n### Install whylogs\n\nInstall whylogs using the pip package manager by running:\n\n```\npip install whylogs\n```\n\n### Log some data\n\nProfiling dataframes, Python dicts, or images is as easy as running:\n\n```python\nfrom whylogs import get_or_create_session\nimport pandas as pd\n\nsession = get_or_create_session()\n\nwith session.logger(dataset_name="my_dataset") as logger:\n    logger.log_dataframe(df)\n```\n\n## Table of Contents\n\n- [whylogs Profiles](#whylogs-profiles)\n- [Visualizing Profiles](#visualizing-profiles)\n- [Features](#features)\n- [Data Types](#data-types)\n- [Integrations](#integrations)\n- [Examples](#examples)\n- [Community](#community)\n- [Roadmap](#roadmap)\n- [Contribute](#contribute)\n\n## whylogs Profiles<a name="whylogs-profiles" />\n\n### What are profiles\n\nwhylogs profiles are the core of the whylogs library. They capture key statistical properties of data, such as the\ndistribution (far beyond simple mean, median, and standard deviation measures), the number of missing values, and a wide\nrange of configurable custom metrics. By capturing these summary statistics, we are able to accurately represent the\ndata and enable all of the use cases described in the introduction.\n\nwhylogs profiles have three properties that make them ideal for data logging: they are **descriptive**, **lightweight**,\nand **mergeable**.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/156462388-ceb4cabd-8a5d-4885-8889-fe393a7b2527.png">\n\n**Descriptive**: whylogs profiles describe the dataset that they represent. This high fidelity representation of\ndatasets is what enables whylogs profiles to be effective snapshots of the data. They are better at capturing the\ncharacteristics of a dataset than a sample would be, as discussed in\nour [Data Logging: Sampling versus Profiling](https://whylabs.ai/blog/posts/data-logging-sampling-versus-profiling) blog\npost.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/156462585-6d679747-c9c6-4724-8640-d85fc70794ca.png">\n\n**Lightweight**: In addition to being a high fidelity representation of data, whylogs profiles also have high\ninformation density. You can easily profile terabytes or even petabytes of data in profiles that are only megabytes\nlarge. Because whylogs profiles are lightweight, they are very inexpensive to store, transport, and interact with.\n\n<br />\n\n<img align="left" src="https://user-images.githubusercontent.com/7946482/156462686-5320701b-dbb5-4981-af7f-28f8c25e6738.png">\n\n**Mergeable**: One of the most powerful features of whylogs profiles is their mergeability. Mergeability means that\nwhylogs profiles can be combined together to form new profiles which represent the aggregate of their constituent\nprofiles. This enables logging for distributed and streaming systems, and allows users to view aggregated data across\nany time granularity.\n\n### How do you generate profiles\n\nOnce whylogs is installed, it\'s easy to generate profiles in both Python and Java environments.\n\nTo generate a profile from a Pandas dataframe in Python, simply run:\n\n```python\nfrom whylogs import get_or_create_session\nimport pandas as pd\n\nsession = get_or_create_session()\n\ndf = pd.read_csv("path/to/file.csv")\n\nwith session.logger(dataset_name="my_dataset") as logger:\n    logger.log_dataframe(df)\n```\n\nFor images, replace `logger.log_dataframe(df)` with `logger.log_image("path/to/image.png")`. Similarly, you can profile\nPython dicts by running `logger.log(dict)` directly on the dict object.\n\n### What can you do with profiles\n\nOnce you’ve generated whylogs profiles, a few things can be done with them:\n\nIn your local Python environment, you can set data constraints or visualize your profiles. Setting data constraints on\nyour profiles allows you to get notified when your data don’t match your expectations, allowing you to do data unit\ntesting and some baseline data monitoring. With the Profile Visualizer, you can visually explore your data, allowing you\nto understand it and ensure that your ML models are ready for production.\n\nIn addition, you can send whylogs profiles to the SaaS ML monitoring and AI observability platform [WhyLabs](whylabs.ai)\n. With WhyLabs, you can automatically set up monitoring for your machine learning models, getting notified on both data\nquality and data change issues (such as data drift). If you’re interested in trying out WhyLabs, check out the always\nfree [Starter edition](whylabs.ai/free), which allows you to experience the entire platform’s capabilities with no\ncredit card required.\n\n## Visualizing Profiles<a name="visualizing-profiles" />\n\n### Multiple profile plots\n\nTo view your logger profiles you can use, methods within `whylogs.viz`:\n\n```python\nvizualization = ProfileVisualizer()\nvizualization.set_profiles([profile_day_1, profile_day_2])\nfigure = vizualization.plot_distribution("<feature_name>")\nfigure.savefig("/my/image/path.png")\n```\n\nIndividual profiles can be saved to local disk, uploaded to AWS S3, or sent to the WhyLabs AI Observability Platform via\nthe WhyLabs API. This can be set up to happen automatically when loggers are closed, per the configuration found in the\nSession configuration.\n\nCurrent profiles from active loggers can be loaded from memory with:\n\n```python\nprofile = logger.profile()\n```\n\n### Profile viewer\n\nYou can also load a local profile viewer, where you upload the `json` summary file. The default path for the json files\nis set as `output/{dataset_name}/{session_id}/json/dataset_profile.json`.\n\n```python\nfrom whylogs.viz import profile_viewer\n\nprofile_viewer()\n```\n\nThis will open a viewer on your default browser where you can load a profile json summary, using\nthe `Select JSON profile` button:\nOnce the json is selected you can view your profile\'s features and associated statistics.\n\n<img src="https://whylabs-public.s3-us-west-2.amazonaws.com/assets/whylogs-viewer.gif" title="whylogs HTML viewer demo">\n\n## Features\n\nwhylogs collects approximate statistics and sketches of data on a column-basis into a statistical profile. These metrics\ninclude:\n\n- Simple counters: boolean, null values, data types.\n- Summary statistics: sum, min, max, median, variance.\n- Unique value counter or cardinality: tracks an approximate unique value of your feature using HyperLogLog algorithm.\n- Histograms for numerical features. whyLogs binary output can be queried to with dynamic binning based on the shape of\n  your data.\n- Top frequent items (default is 128). Note that this configuration affects the memory footprint, especially for text\n  features.\n\nSome other key features about whylogs:\n\n- Accurate data profiling: whylogs calculates statistics from 100% of the data, never requiring sampling, ensuring an\n  accurate representation of data distributions\n- Lightweight runtime: whylogs utilizes approximate statistical methods to achieve minimal memory footprint that scales\n  with the number of features in the data\n- Any architecture: whylogs scales with your system, from local development mode to live production systems in\n  multi-node clusters, and works well with batch and streaming architectures\n- Configuration-free: whylogs infers the schema of the data, requiring zero manual configuration to get started\n- Tiny storage footprint: whylogs turns data batches and streams into statistical fingerprints, 10-100MB uncompressed\n- Unlimited metrics: whylogs collects all possible statistical metrics about structured or unstructured data\n\n## Data Types<a name="data-types" />\n\nwhylogs supports both structured and unstructured data, specifically:\n\n| Data type       | Features                                                  | Notebook Example                                                                                                              |\n| --------------- | --------------------------------------------------------- | ----------------------------------------------------------------------------------------------------------------------------- |\n| Structured Data | Distribution, cardinality, schema, counts, missing values | [Getting started with structured data](https://github.com/whylabs/whylogs-examples/blob/mainline/python/GettingStarted.ipynb) |\n| Images          | exif metadata, derived pixels features, bounding boxes    | [Getting started with images](https://github.com/whylabs/whylogs-examples/blob/mainline/python/Logging_Images.ipynb)          |\n| Video           | In development                                            | [Github Issue #214](https://github.com/whylabs/whylogs/issues/214)                                                            |\n| Tensors         | derived 1d features (more in developement)                | [Github Issue #216](https://github.com/whylabs/whylogs/issues/216)                                                            |\n| Text            | top k values, counts, cardinality                         | [String Features](https://github.com/whylabs/whylogs/blob/mainline/examples/String_Features.ipynb)                            |\n| Audio           | In developement                                           | [Github Issue #212](https://github.com/whylabs/whylogs/issues/212)                                                            |\n\n## Integrations\n\n![current integration](images/integrations.001.png)\n| Integration | Features | Resources | | --- | --- | --- | | Spark | Run whylogs in Apache Spark\nenvironment| <ul><li>[Code Example](https://github.com/whylabs/whylogs-examples/blob/mainline/scala/src/main/scala/WhyLogsDemo.scala)</li></ul>\n| | Pandas | Log and monitor any pandas dataframe\n| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/blob/mainline/python/logging_example.ipynb)</li><li>[whylogs:\nEmbrace Data Logging](https://whylabs.ai/blog/posts/whylogs-embrace-data-logging)</li></ul> | | Kafka | Log and monitor\nKafka topics with\nwhylogs| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/blob/mainline/python/Kafka.ipynb)</li><li> [Integrating whylogs into your Kafka ML Pipeline](https://whylabs.ai/blog/posts/integrating-whylogs-into-your-kafka-ml-pipeline) </li></ul>\n| | MLflow | Enhance MLflow metrics with whylogs:\n| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/blob/mainline/python/MLFlow%20Integration%20Example.ipynb)</li><li>[Streamlining\ndata monitoring with whylogs and MLflow](https://whylabs.ai/blog/posts/on-model-lifecycle-and-monitoring)</li></ul> | |\nGithub actions | Unit test data with whylogs and github\nactions| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/tree/mainline/github-actions)</li></ul>\n| | RAPIDS | Use whylogs in RAPIDS environment\n| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/blob/mainline/python/RAPIDS%20GPU%20Integration%20Example.ipynb)</li><li>[Monitoring\nHigh-Performance Machine Learning Models with RAPIDS and\nwhylogs](https://whylabs.ai/blog/posts/monitoring-high-performance-machine-learning-models-with-rapids-and-whylogs)</li></ul>\n| | Java | Run whylogs in Java\nenvironment| <ul><li>[Notebook Example](https://github.com/whylabs/whylogs-examples/blob/mainline/java/demo1/src/main/java/com/whylogs/examples/WhyLogsDemo.java)</li></ul>\n| | Docker | Run whylogs as in Docker\n| <ul><li>[Rest Container](https://docs.whylabs.ai/docs/integrations-rest-container)</li></ul>|\n| AWS S3 | Store whylogs profiles in S3\n| <ul><li>[S3 example](https://github.com/whylabs/whylogs-examples/blob/mainline/python/S3%20example.ipynb)</li></ul>\n\n## Examples\n\nFor a full set of our examples, please check out the [examples folder](/examples/).\n\nCheck out our example notebooks with\nBinder: [![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/whylabs/whylogs-examples/HEAD)\n\n- [Getting Started notebook](https://github.com/whylabs/whylogs-examples/blob/mainline/python/GettingStarted.ipynb)\n- [Logging Example notebook](https://github.com/whylabs/whylogs-examples/blob/mainline/python/logging_example.ipynb)\n- [Logging Images](https://github.com/whylabs/whylogs-examples/blob/mainline/python/Logging_Images.ipynb)\n- [MLflow Integration](https://github.com/whylabs/whylogs-examples/blob/mainline/python/MLFlow%20Integration%20Example.ipynb)\n\n## Roadmap\n\nwhylogs is maintained by [WhyLabs](https://whylabs.ai).\n\n## Community\n\nIf you have any questions, comments, or just want to hang out with us, please\njoin [our Slack channel](http://join.slack.whylabs.ai/).\n\n## Contribute\n\n### How to Contribute\n\nWe welcome contributions to whylogs. Please see\nour [contribution guide](https://github.com/whylabs/whylogs/blob/mainline/CONTRIBUTING.md) and\nour [development guide](https://github.com/whylabs/whylogs/blob/mainline/DEVELOPMENT.md) for details.\n\n### Contributors\n\n<a href="https://github.com/whylabs/whylogs/graphs/contributors">\n  <img src="https://contrib.rocks/image?repo=whylabs/whylogs" />\n</a>\n\nMade with [contrib.rocks](https://contrib.rocks).\n',
    'author': 'WhyLabs.ai',
    'author_email': 'support@whylabs.ai',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://docs.whylabs.ai',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7.1,<4.0',
}


setup(**setup_kwargs)



from whylogs.core.constraints import SummaryConstraint
from whylogs.core.proto.v0 import Op

MAX_SET_DISPLAY_MESSAGE_LENGTH = 20


def maxLessThanEqualConstraint(
    value: float = None, field: str = None, name: str = None, verbose: bool = False
) -> SummaryConstraint:
    """
    Defines a summary constraint on the maximum value of a feature. The maximum can be defined to be
    less than or equal to some value,
    or less than or equal to the values of another summary field of the same feature, such as the mean (average).

    Parameters
    ----------
    value : numeric (one-of)
        Represents the value which should be compared to the maximum value of the specified feature,
        for checking the less than or equal to constraint.
        Only one of `value` and `field` should be supplied.
    field : str (one-of)
        The field is a string representing a summary field
        e.g. `min`, `mean`, `max`, `stddev`, etc., for which the value will be used for
        checking the less than or equal to constraint.
        Only one of `field` and `value` should be supplied.
    name : str
        Name of the constraint used for reporting
    verbose : bool
        If true, log every application of this constraint that fails.
        Useful to identify specific streaming values that fail the constraint.

    Returns
    -------
        SummaryConstraint -  a summary constraint defining a constraint on the maximum value to be less than
        or equal to some value / summary field

    """
    if name is None:
        name = f"maximum is less than or equal to {value}"

    return SummaryConstraint("max", Op.LE, value=value, second_field=field, name=name, verbose=verbose)
